#!/usr/bin/env python3

# Settings -> Project: papr -> Project Structure
#   use only papr as "Sources"

import sys

from .lib.cmd_default import cmd_default
from .lib.cmd_init import cmd_init
from .lib.cmd_last import cmd_last
from .lib.cmd_list import cmd_list
from .lib.cmd_read import cmd_read
from .lib.cmd_search import cmd_search
from .lib.ui import run_ui
from .lib.config import Config
from .lib.repository import Repository
from .lib.help import help
from .lib.cmd_add import cmd_add
from .lib.cmd_fetch import cmd_fetch


def parse_command(conf: Config, repo: Repository) -> None:
    if len(sys.argv) > 1:
        c = sys.argv[1]
        if c == "init":
            cmd_init(repo)
            return
        elif c == "help" or c == "--help" or c == "-h":
            help(0)
        elif c == "default":
            cmd_default(conf, repo)
            return

    # For all other commands we need a valid repository.
    if not repo.is_valid():
        print("No repository.")
        sys.exit(1)

    # no arguments given -> show UI
    if len(sys.argv) < 2:
        run_ui(sys.argv[2:], repo)
        sys.exit(0)

    c = sys.argv[1]
    if c == "list":
        cmd_list(repo)
    elif c == "fetch":
        cmd_fetch(sys.argv[2:], repo)
    elif c == "read":
        cmd_read(sys.argv[2:], repo)
    elif c == "search":
        cmd_search(sys.argv[2:], repo)
    elif c == "last":
        cmd_last(sys.argv[2:], repo)
    elif c == "add":
        cmd_add(sys.argv[2:], repo)
    else:
        help(1)


def main() -> None:
    # Reads the configuration from "~/.papr/". Creates a default config if
    # it doesn't exist yet.
    conf = Config()
    # If we are in a repository create an instance of this repository. If we
    # are not in a repository create an instance of the default repository. If
    # there is no default repository, Repository.is_valid() will return False.
    r = Repository(conf)
    parse_command(conf, r)


if __name__ == "__main__":
    main()



from pathlib import Path
import json
import os


class Config:
    def __init__(self):
        self.config_path = str(Path.home()) + "/.papr"
        self.config_file = self.config_path + "/papr.cfg"
        self._read_config()

    def get(self, key):
        return self.config[key]

    def _read_config(self):
        if not os.path.exists(self.config_path):
            os.mkdir(self.config_path)
        if not os.path.exists(self.config_file):
            self._create_default()
        f = open(self.config_file, "r")
        self.config = json.loads(f.read())
        f.close()

    def _create_default(self):
        self.config = {"cfg_version": "0.0.1", "default_repo": "null"}
        self._write_config()

    def _write_config(self):
        f = open(self.config_file, "w")
        f.write(json.dumps(self.config))
        f.close()

    def set_default_repo(self, path):
        self.config["default_repo"] = path
        self._write_config()




import sys

from .tools import show_pdf


def cmd_read(args, repo):
    if len(args) == 0:
        print("You need to specify a paper id.")
        sys.exit(1)
    p = repo.get_paper(int(args[0]))
    if not p:
        print("Not found.")
        sys.exit(1)
    show_pdf(p, repo.pdf_path())





import sys
import termcolor

from .console import cursor_on, cursor_up, cursor_off
from .edit import notes_of_paper, tags_of_paper
from .termin import read_key
from .termout import rows, empty_line, print_paper, print_header, cols
from .tools import filter_list, show_pdf
from .cmd_fetch import NEWTAG


def colwidth(s: str):
    c = cols()
    return s + (" " * (c - len(s)))


def run_ui(args, repo):
    r = repo.list()
    if len(r) == 0:
        print("Repository is empty.")
        sys.exit(0)

    if len(args) > 0:
        r = filter_list(r, args[0])

    print(termcolor.colored(" ", "white", attrs=["bold"]))
    print_header()
    m = rows() - 4 - 1             # have added -1 as otherwise the header flickers sometimes when scrolling
    n = len(r)
    window_rows = min(m, n)        # number of rows of the view
    view = n - window_rows         # index in r of the first element in the view
    selected = window_rows - 1     # row selected within the view
    initial_window_rows = window_rows
    search = ""
    in_search = False
    cursor_off()
    papers = r[:]
    while True:
        cursor_up(4)
        if in_search:
            print(termcolor.colored(colwidth("ESC: cancel + back to selection mode | ENTER: back to selection mode"), "white", "on_blue", attrs=["bold"]))
        else:
            print(termcolor.colored(colwidth("ESC/q: quit | ENTER: open | ↑: up | ↓: down | s: search | n: notes | t: tags"),
                                    "white", attrs=["bold"]))
        print_header()

        cnt = 0
        for idx, p in enumerate(papers[view:view + m]):
            print_paper(p, idx == selected)
            cnt += 1
        while cnt < initial_window_rows:
            cnt += 1
            print(empty_line())

        # search line
        sys.stdout.write(empty_line() + "\r")
        if in_search or len(search) > 0:
            sys.stdout.write("search: " + search + "▃")
            sys.stdout.flush()

        k = read_key()
        if k is None:
            cursor_up(initial_window_rows + 1)
            continue
        if not in_search:
            if k == 'i' or k == '#':
                if selected == 0:
                    view = max(view - 1, 0)
                else:
                    selected -= 1
                cursor_up(initial_window_rows + 1)
            elif k == 'k' or k == '+':
                if selected == window_rows - 1:
                    if selected + view + 1 < n:
                        view += 1
                    pass
                else:
                    selected += 1
                cursor_up(initial_window_rows + 1)
            elif k == 'n':
                notes_of_paper(repo, papers[view + selected])
                cursor_up(initial_window_rows + 1)
            elif k == 't':
                tags_of_paper(repo, papers[view + selected])
                cursor_up(initial_window_rows + 1)
            elif ord(k) == 10:
                papers[view + selected].remove_tag(NEWTAG)
                repo.update_paper(papers[view + selected])
                show_pdf(papers[view + selected], repo.pdf_path())
                cursor_up(initial_window_rows + 1)
            elif ord(k) == 27 or k == 'q':
                break
            elif k == 's':
                cursor_up(initial_window_rows + 1)
                in_search = True
            elif k >= '0' and k <= '5':
                stars = int(k)
                papers[view + selected].set_stars(stars)
                repo.update_paper(papers[view + selected])
                cursor_up(initial_window_rows + 1)
            else:
                cursor_up(initial_window_rows + 1)
        else:
            update_search = False
            if ord(k) == 27:
                in_search = False
                search = ""
                cursor_up(initial_window_rows + 1)
                update_search = True
            elif ord(k) == 10:
                in_search= False
                cursor_up(initial_window_rows + 1)
            elif ord(k) == 127:
                search = search[:-1]
                cursor_up(initial_window_rows + 1)
                update_search = True
            else:
                search += k
                cursor_up(initial_window_rows + 1)
                update_search = True
            if update_search:
                if len(search) == 0:
                    papers = r[:]
                else:
                    papers = filter_list(r, search)
                m = rows() - 4 - 1
                n = len(papers)
                window_rows = min(m, n)        # number of rows of the view
                view = n - window_rows         # index in r of the first element in the view
                selected = window_rows - 1     # row selected within the view

    cursor_on()
    print()





#!/usr/bin/env python3

import os
import unittest
from pathlib import Path
from lib.config import Config
import shutil


def get_config_path():
    return str(Path.home()) + "/.papr"


def get_config_file():
    return str(Path.home()) + "/.papr/papr.cfg"


def create_default_config():
    shutil.rmtree(get_config_path())
    Config()


class TestConfig(unittest.TestCase):

    # ------- HELP FUNCTIONS ----------------------------------------


    # --------- TESTS -----------------------------------------------

    def test_config_does_not_exist(self):
        """
        Pre: ~/.papr and ~/.papr/papr.cfg does not exist.
        Check:
        - configuratino file ~/papr/papr.cfg should exist
        - default reposiroty = "null"
        - cfg_version = "0.0.1"
        :return:
        """
        p = get_config_path()
        f = get_config_file()

        self.assertFalse(os.path.exists(p))

        c = Config()
        self.assertEqual(c.get("cfg_version"), "0.0.1")
        self.assertEqual(c.get("default_repo"), "null")
        self.assertTrue(os.path.exists(p))
        self.assertTrue(os.path.exists(f))

    def test_config_exists(self):
        """
        Pre: ~/.papr/papr.cfg exists
        Check:
        :return:
        """
        create_default_config()
        self.assertTrue(os.path.exists(get_config_file()))
        c = Config()
        self.assertEqual(c.get("cfg_version"), "0.0.1")
        self.assertEqual(c.get("default_repo"), "null")


if __name__ == "__main__":
    unittest.main()



import sys


def cursor_up(n):
    # http://tldp.org/HOWTO/Bash-Prompt-HOWTO/x361.html
    print('\033[' + str(n) + 'A')


def cursor_off():
    sys.stdout.write("\x1b[?25l")
    sys.stdout.flush()


def cursor_on():
    sys.stdout.write("\x1b[?25h")
    sys.stdout.flush()



import json


class Paper:
    def __init__(self, idx, filename, title, tags=[]):
        self._idx = idx
        self._filename = filename
        self._title = title
        self._msg = ""
        self._tags = tags
        self._stars = 0

    @staticmethod
    def from_json(idx, jsonstr):
        p = Paper(idx, filename=None, title=None)
        p._idx = idx
        data = json.loads(jsonstr)
        p._filename = data["filename"]
        p._title = data["title"]
        p._msg = data.get("msg", "")
        p._tags = [i for i in data.get("tags", "").split(",") if len(i) > 0]
        p._stars = int(data.get("stars", "0"))
        return p

    def as_json(self):
        d = {
            "filename": self._filename,
            "title": self._title,
            "msg": self._msg,
            "tags": ",".join(self._tags),
            "stars": self._stars
        }
        return json.dumps(d)

    def remove_tag(self, tag):
        self._tags = [i for i in self._tags if i != tag]

    def idx(self):
        return self._idx

    def filename(self):
        return self._filename

    def title(self):
        return self._title

    def msg(self):
        return self._msg

    def update_msg(self, msg):
        self._msg = msg

    def has_notes(self):
        return len(self._msg) > 0

    def tags(self):
        return sorted(self._tags)

    def set_tags(self, tags):
        self._tags = tags

    def stars(self):
        return self._stars

    def set_stars(self, n):
        self._stars = max(0, min(5, n))



import os
import termios
import sys
from select import select


def read_key():
    fd = sys.stdin.fileno()
    oldterm = termios.tcgetattr(fd)
    newattr = termios.tcgetattr(fd)
    newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
    termios.tcsetattr(fd, termios.TCSANOW, newattr)
    c = []
    try:
        select([fd], [], [])
        c = [i for i in os.read(fd, 50)]
    except IOError:
        pass
    termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
    if len(c) == 1:
        return chr(c[0])
    else:
        if c == [27, 91, 65]:
            return '#'
        elif c == [27, 91, 66]:
            return '+'



import re
from subprocess import Popen, DEVNULL

from .paper import Paper


VIEWER = "/usr/bin/evince"


def exists_in(p, q):
    try:
        r = re.compile(q.lower())
        if r.search(p.title().lower()):
            return True
        return False
    except Exception:
        return False


def filter_list(l, query):
    return [i for i in l if exists_in(i, query)]


def show_pdf(p: Paper, repo_path):
    Popen([VIEWER, repo_path + "/" + p.filename()], stderr=DEVNULL, stdout=DEVNULL)





import sys

from .config import Config


def cmd_default(conf: Config, repo) -> None:
    if not repo.is_local_repository():
        print("You are not in a repository.")
        sys.exit(1)

    conf.set_default_repo(repo.pdf_path())





import os
import termcolor


def _hr():
    cols, _ = os.get_terminal_size(0)
    print("─" * cols)


def rows():
    _, rows = os.get_terminal_size(0)
    return rows


def cols():
    cols, _ = os.get_terminal_size(0)
    return cols


def empty_line():
    cols, _ = os.get_terminal_size(0)
    return " " * cols


def print_header():
    cols, _ = os.get_terminal_size(0)
    print("Id    Title " + " " * (cols - 12 - 8) + "Stars" + "  ")
    _hr()
    return cols


def print_paper(paper, selected = False):
    cols, _ = os.get_terminal_size(0)
    # 5 columns for id
    # 1 columns for space
    # 2 columns at the end of the line for notes
    # n columns for filename

    stars = ""
    if paper.stars() >= 0:
        stars = " " * (5 - paper.stars()) + "*" * paper.stars()
    len_stars = len(stars) + 3

    tags = ""
    len_tags = 0
    if len(paper.tags()) > 0:
        tags = ",".join(paper.tags())
        len_tags = len(tags) + 3

    n = cols - (5 + 1 + 2) - len_stars - len_tags
    f = paper.title()
    if len(f) > n:
        f = f[:n - 3] + "..."

    m = "{:5d} {}".format(paper.idx(), f)
    s = colored("line", selected, m)
    if len(m) < cols:
        s = s + colored("line", selected, " " * (cols - len(m) - 2 - len_stars - len_tags))

    if len_tags > 0:
        s = s + colored("line", selected, " [") + colored("tags", selected, tags) + colored("line", selected, "]")
    if len_stars > 0:
        s = s + colored("line", selected, " [" + stars + "]")

    if paper.has_notes():
        s += colored("line", selected, " ✺")
    else:
        s += colored("line", selected, "  ")

    print(s)


def colored(typ, selected, s):
    mapping = {"line": "white", "tags": "green"}
    if selected:
        return termcolor.colored(s, mapping[typ], 'on_red', attrs=["bold"])
    else:
        return termcolor.colored(s, mapping[typ])


def print_papers(papers):
    for paper in papers:
        print_paper(paper)




import math
import sys
import urllib
import urllib.request
import os
import shutil
import re
from bs4 import BeautifulSoup

from .paper import Paper
from .repository import Repository


NEWTAG="unread"


def title_as_filename(s):
    return re.sub(r'[^a-z0-9]', "_", re.sub(r'\s+', "_", s.lower()))


def normalize_title(s):
    return re.sub(r'\s+', " ", s.replace("\n", " "))


def prepare_data(title: str, repo: Repository):
    title = normalize_title(title)
    idx = repo.next_id()
    filename = "{:05d}_{}.pdf".format(idx, title_as_filename(title))
    abspath = repo.pdf_path() + "/" + filename
    return title, idx, filename, abspath


def add_local_file(f: str, args, repo: Repository):
    if len(args) == 0:
        print("For local files you need to specify the title of the paper.")
        sys.exit(1)

    title, idx, filename, abspath = prepare_data(args[0], repo)
    if not os.path.exists(abspath):
        shutil.copy(f, abspath)
    p = Paper(idx=idx, filename=filename, title=title, tags=[NEWTAG])
    repo.add_paper(p)
    print("Added paper.")
    print("Title:", title)
    print("Filename:", filename)


def is_text_or_html(s):
    return s is not None and (s.lower().find("text") >= 0 or s.lower().find("html") >= 0)


def bar(p):
    chars = "▏▎▍▌▋▊▉█"
    N = 40
    l = p * N / 100.0
    todo = "." * (N - int(l))
    done = "█" * int(l)

    f = l - math.floor(l)
    if f > 0:
        todo = "." * (N - int(l) - 1)
        done = done + chars[int(f * len(chars))]

    r = str(int(255 - 255 * p / 100.0))
    g = str(int(255 * p / 100.0))
    b = "0"
    sys.stdout.write("\r|\x1b[38;2;{};{};{}m{}\x1b[0m{}| {:3d}% ".format(r, g, b, done, todo, int(p)))
    sys.stdout.flush()


def download(rsp, no_progress_bar=False):
    n = rsp.getheader("content-length")
    data = bytes()
    if n is not None:
        n = int(n)
        s = n
        while n > 0:
            t = rsp.read(min(1024, n))
            data += t
            n -= min(1024, n)
            if not no_progress_bar:
                bar(100.0 * (s - n) / s)
        if not no_progress_bar:
            print()
    else:
        chars = "|/-|\\"
        n = 0
        while True:
            t = rsp.read(10240)
            if len(t) == 0:
                break
            data += t
            if not no_progress_bar:
                sys.stdout.write("\rloading " + chars[n % len(chars)])
                sys.stdout.flush()
            n += 1
        if not no_progress_bar:
            print("\r          ")
    return data


def parse_page(data):
    h = BeautifulSoup(data, "html.parser")
    r = h.find_all("meta")
    title = [i.get("content") for i in r if i.get("name") == "citation_title"]
    pdfurl = [i.get("content") for i in r if i.get("name") == "citation_pdf_url"]
    if len(title) > 0 and len(pdfurl) > 0:
        return title[0], pdfurl[0]
    return None, None


def do_fetch_download(repo: Repository, url, title):
    req = urllib.request.Request(url)
    rsp = urllib.request.urlopen(req)
    typ = rsp.getheader("content-type")
    if not is_text_or_html(typ) and title == "":  # for PDF files we need a title
        print("The response is not html. You need to specify a title.", file=sys.stderr)
        sys.exit(1)
    data = download(rsp, is_text_or_html(typ))
    if is_text_or_html(typ):
        t, pdfurl = parse_page(data)
        if t is None:
            print("Error")
            sys.exit(1)
        title = t
        req = urllib.request.Request(pdfurl)
        rsp = urllib.request.urlopen(req)
        data = download(rsp)

    title, idx, filename, abspath = prepare_data(title, repo)

    k = open(abspath, "w")
    k.buffer.write(data)
    k.close()

    p = Paper(idx=idx, filename=filename, title=title, tags=[NEWTAG])
    repo.add_paper(p)
    print("   Title:", title)
    print("Filename:", filename)
    print("Added paper.")
    print()


def do_fetch(args, repo: Repository):
    title = ""
    if len(args) >= 2:
        title = args[1]
    fname = None
    if len(args) >= 1:
        fname = args[0]
    if os.path.exists(fname):
        add_local_file(fname, args[1:], repo)
    else:
        do_fetch_download(repo, url=fname, title=title)


def cmd_fetch(args, repo: Repository):
    if len(args) == 0:
        print("You need to specify an option, filename or URL.")
        sys.exit(1)

    if args[0] == "--urls":
        if len(args) < 2:
            print("You need to specify a file.", file=sys.stderr)
            sys.exit(1)
        with open(args[1], "r") as f:
            for line in f:
                s = line.strip()
                if len(s) > 0:
                    print("fetching", s, "...", file=sys.stderr)
                    do_fetch([s], repo)
    else:
        do_fetch(args, repo)





import tempfile
import os
from subprocess import call

from .paper import Paper
from .repository import Repository


def editor(msg, n=None):
    e = os.getenv('EDITOR', 'vim')
    _, pth = tempfile.mkstemp(".tmp")
    f = open(pth, "wt")
    f.write(msg)
    f.close()
    cmd = [e]
    if n is not None and e == "vim":
        cmd.append("+" + str(n))
    cmd.append(pth)
    call(cmd)
    f = open(pth, "r")
    msg = f.read()
    f.close()
    return msg


def notes_of_paper(repo: Repository, p: Paper):
    msg = p.msg()
    msg = editor(msg, 1).strip()
    p.update_msg(msg)
    repo.update_paper(p)


def tags_of_paper(repo: Repository, p: Paper):
    msg = "COMMA SEPARATED LIST OF WORDS\n" + ",".join(p.tags())
    msg = editor(msg, 2)
    pos = msg.find("\n")
    if pos >= 0:
        msg = msg[pos+1:]
    msg = msg.replace("\n", ",")
    tags = [j for j in [i.strip().lower() for i in msg.split(",")] if len(j) > 0]
    p.set_tags(tags)
    repo.update_paper(p)



import setuptools
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="papr",
    version="0.0.13",
    entry_points = {
        "console_scripts": ['papr = papr.cli:main']
        },
    author="Daniel Etzold",
    author_email="detzold@gmail.com",
    description="A command line tool to manage scientific papers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/daniel-e/papr",
    packages=['papr', 'papr/lib'],
    #include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'bs4',
        'termcolor'
    ]
)



import os
import sys

from .db import Db
from .config import Config
from .paper import Paper

REPO_META = ".paper"


class Repository:
    def __init__(self, config: Config):
        self.repo_pdf_path = None
        self.repo_meta_path = None
        self.local = None
        self.db = None
        self.config = config
        self._init_paths()
        if self.is_valid():
            self.db = Db(self.repo_meta_path)
            if not self.db.check_version():
                print("Found incompatible version.")
                sys.exit(1)

    def init(self):
        """
        Creates directories, databases and updates configuration but does not change
        this isntance into a valid repository.
        :return:
        """
        # Update configuration: set default repository.
        self.config.set_default_repo(os.getcwd())
        # Create database.
        os.mkdir(os.getcwd() + "/" + REPO_META)
        self.db = Db.create(os.getcwd() + "/" + REPO_META)

    def is_local_repository(self):
        return self.local

    def is_valid(self):
        """
        :return: True if this object represents a valid repository.
        """
        return self.repo_meta_path is not None and self.repo_pdf_path is not None

    def pdf_path(self):
        return self.repo_pdf_path

    def _init_paths(self):
        # First check if the current working directory is a repository.
        self.repo_pdf_path = os.getcwd()  # $PWD
        self.repo_meta_path = self.repo_pdf_path + "/" + REPO_META  # $PWD/.paper/
        self.local = True

        if not os.path.exists(self.repo_meta_path):
            # If it does not exist read the location of the repository from
            # the configuration.
            self.repo_pdf_path = self.config.get("default_repo")
            self.repo_meta_path = self.repo_pdf_path + "/" + REPO_META
            self.local = False

        if not os.path.exists(self.repo_meta_path):
            self.repo_pdf_path = None
            self.repo_meta_path = None
            self.local = None

    def list(self):
        return self.db.list()

    def next_id(self):
        return self.db.next_id()

    def add_paper(self, p: Paper):
        return self.db.add_paper(p)

    def get_paper(self, idx):
        return self.db.get(idx)

    def update_paper(self, p):
        self.db.update_paper(p)
